#define MIN_CLUES 31

var
	generating = 0
	gridminx; gridmaxx; gridminy; gridmaxy


mob/verb/generate(threshold as num)
	set desc = "Enter approximate number of clues (81 for full board)."

	if(threshold < MIN_CLUES)
		world << "You must choose at least [MIN_CLUES] clues."
		return

	if(generating)
		world << "Can't interrupt generation in progress."
		return

	var/mapgen/MG = new()


	world << "Generating approx. [threshold] clues..."
	sleep(1)

	MG.MakeMap(81 - threshold)


world
	turf = /turf/background


	New()
		. = ..()
		spawn(1)
			world << "Welcome to Oki-Doku: The world's cheesiest Number Place game!\n\n" + \
			"Instructions: When you are certain of a cell's contents, click it. If you don't know how to play Sudoku, see Google.com.\n"


mapgen
	var/list/unassigned = list()


	proc/MakeMap(threshold)
		generating = 1

		while(1)
			if(Init()) break

		while(1)
			PlaceBlanks(threshold)

			if(IsSolvable()) break

		world << "Done!\n"

		generating = 0


	proc/IsSolvable()
		var/list
			unsolved = list()
			doblank = list()
		var/turf/cell
			T
			neighbor

		for(T in world)
			if(T.number.icon_state == "?")
				unsolved += T
				T.possible = list(1, 2, 3, 4, 5, 6, 7, 8, 9)
			else
				T.possible = list(T.value)

		while(unsolved.len)
			var/foundone = 0

			for(T in unsolved)
				for(neighbor in T.GetRow())
					if(neighbor.number.icon_state != "?")
						T.possible -= neighbor.value

				for(neighbor in T.GetColumn())
					if(neighbor.number.icon_state != "?")
						T.possible -= neighbor.value

				for(neighbor in T.GetBox())
					if(neighbor.number.icon_state != "?")
						T.possible -= neighbor.value

			for(T in unsolved)
				if(length(T.possible) == 1)
					foundone = 1
					T.number.icon_state = num2text(T.value)
					unsolved -= T
					doblank += T

			if(!foundone) return 0

		for(T in doblank)
			T.number.icon_state = "?"

		return 1


	proc/PlaceBlanks(threshold)
		var/list/allcells = list()
		var/list/blanks = list()
		var/turf/cell/T

		for(T in world)
			allcells += T
			T.number.icon_state = num2text(T.value)

		while(length(blanks) < threshold)
			while(1)
				T = pick(allcells)
				if(T.number.icon_state != "?") break

			T.number.icon_state = "?"
			blanks += T
			T = GetComplement(T)
			T.number.icon_state = "?"
			blanks += T

		for(T in allcells)
			if(T.number.icon_state != "?")
				T.number.icon_state = num2text(T.value)


	proc/GetComplement(turf/cell/T)
		var
			compx; compy

		compx = gridmaxx - (T.x - gridminx)
		compy = gridmaxy - (T.y - gridminy)
		return locate(compx, compy, 1)


	proc/Init()
		var/turf/cell
			C
			worst

		unassigned.len = 0

		for(C in world)
			C.value = 0
			C.number.icon_state = null
			unassigned += C
			C.possible = list(1, 2, 3, 4, 5, 6, 7, 8, 9)

		while(length(unassigned))
			worst = pick(unassigned)
			for(C in unassigned)
				if(length(C.possible) < length(worst.possible))
					worst = C

			if(!AssignValue(worst)) return 0

		return 1


	proc/AssignValue(turf/cell/C)
		if(!length(C.possible)) return 0

		C.value = pick(C.possible)
		//C.number.icon_state = num2text(C.value)
		unassigned -= C
		C.possible.len = 0

		var/turf/cell/T
		for(T in C.GetRow())
			T.possible -= C.value

		for(T in C.GetColumn())
			T.possible -= C.value

		for(T in C.GetBox())
			T.possible -= C.value

		return 1


turf
	background
		icon = 'background.dmi'


	cell
		icon = 'cell.dmi'
		var/value = 0
		var/obj/number/number
		var/list/possible


		Click()
			number.Click()


		New()
			. = ..()
			number = new(src)


	proc/GetRow()
		return block(locate(gridminx, y, 1), locate(gridmaxx, y, 1)) - src


	proc/GetColumn()
		return block(locate(x, gridminy, 1), locate(x, gridmaxy, 1)) - src


	proc/GetBox()
		var/list/result = list()
		for(var/turf/cell/T in loc)
			if(T != src) result += T
		return result


obj
	number
		icon = 'number.dmi'
		icon_state = ""


		Click()
			icon_state = num2text(loc:value)
			//var/turf/cell/T = loc
			//for(var/k in T.possible) world << k

	border
		icon = 'border.dmi'


area
	box_1


		New()
			gridminx = x
			gridmaxy = y + 2


	box_2


	box_3


	box_4


	box_5


	box_6


	box_7


	box_8


	box_9


		New()
			gridminy = y
			gridmaxx = x + 2




